#include "data.lua"
#include "tdmp/networking.lua"
#include "tdmp/json.lua"
#include "tdmp/player.lua"
#include "tdmp/hooks.lua"


--General functions
local function toEuler(quat)
    local x, y, z, w = quat[1], quat[2], quat[3], quat[4]
    -- Credit to https://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToEuler/index.htm

    local bank, heading, attitude

    local s = 2 * x * y + 2 * z * w
    if s >= 1 then
        heading = 2 * math.atan2(x, w)
        bank = 0
        attitude = math.pi / 2
    elseif s <= -1 then
        heading = -2 * math.atan2(x, w)
        bank = 0
        attitude = math.pi / -2
    else
        bank = math.atan2(2 * x * w - 2 * y * z, 1 - 2 * x ^ 2 - 2 * z ^ 2)
        heading = math.atan2(2 * y * w - 2 * x * z,
                             1 - 2 * y ^ 2 - 2 * z ^ 2)
        attitude = math.asin(s)
    end

    return math.deg(bank), math.deg(heading), math.deg(attitude)
end

local function VecAngle(v1, v2)
    local numerator = VecDot(v1, v2)
    local denominator = math.sqrt((v1[1] * v1[1]) + (v1[2] * v1[2]) + (v1[3] * v1[3])) * math.sqrt((v2[1] * v2[1]) + (v2[2] * v2[2]) + (v2[3] * v2[3]))
    local rad = math.acos(numerator/denominator)

    return math.deg(rad)
end

local function getPlayerDirection()
    local direction = {}

    local ct = GetPlayerCameraTransform()
    local pVec = TransformToParentVec(ct, Vec(0, 0, -1))

    local xzVec = Vec(pVec[1], 0, pVec[3])
    local xBaseVec = Vec(1, 0, 0)

    direction.yaw = VecAngle(xBaseVec, xzVec)
    direction.pitch = VecAngle(xzVec, pVec)

    if(pVec[1] < 0 and pVec[3] < 0 or pVec[1] > 0 and pVec[3] < 0) then
        direction.yaw = (180 - direction.yaw) + 180
    end 

    if(pVec[2] < 0) then
        direction.pitch = -direction.pitch
    end 

    return direction
end

local function getCameraFront()
    local direction = Vec(0, 0, 0)

    yaw = getPlayerDirection().yaw
    pitch = getPlayerDirection().pitch

    direction[1] = math.cos(math.rad(yaw)) * math.cos(math.rad(pitch))
    direction[2] = math.sin(math.rad(pitch))
    direction[3] = math.sin(math.rad(yaw)) * math.cos(math.rad(pitch))

    return VecNormalize(direction)
end

local function has_value(tab, val)
    for index, value in ipairs(tab) do
        if value == val then
            return true
        end
    end
    return false
end

local function find_value(tab, val)
    for index, value in ipairs(tab) do
        if value == val then
            return index
        end
    end
    return false
end

local function progressBar(w, h, t) --common.lua line 44 (edited)
	UiPush()
		UiAlign("left top")
		UiColor(0, 0, 0, 0.5)
		UiImageBox("ui/common/box-solid-10.png", w, h, 6, 6)
		if t > 0 then
			UiTranslate(2, 2)
			w = (w-4)*t
			if w < 12 then w = 12 end
			h = h-4
            UiColor(1,(1-t),(1-(t*2)),1)
			UiImageBox("ui/common/box-solid-6.png", w, h, 6, 6)
		end
	UiPop()
end

local function round(x, n)
    n = math.pow(10, n or 0)
    x = x * n
    if x >= 0 then x = math.floor(x + 0.5) else x = math.ceil(x - 0.5) end
    return x / n
end

local function getBodySpeed(body)
    local trans = GetBodyTransform(body)
    local vel = TransformToLocalVec(trans, GetBodyVelocity(body))
    local speed = math.sqrt(vel[1]*vel[1]+vel[2]*vel[2]+vel[3]*vel[3])
    return speed
end

local function serializeTable(val, name, skipnewlines, depth)
    skipnewlines = skipnewlines or false
    depth = depth or 0

    local tmp = string.rep(" ", depth)

    if name then tmp = tmp .. name .. " = " end

    if type(val) == "table" then
        tmp = tmp .. "{" .. (not skipnewlines and "\n" or "")

        for k, v in pairs(val) do
            tmp =  tmp .. serializeTable(v, k, skipnewlines, depth + 1) .. "," .. (not skipnewlines and "\n" or "")
        end

        tmp = tmp .. string.rep(" ", depth) .. "}"
    elseif type(val) == "number" then
        tmp = tmp .. tostring(val)
    elseif type(val) == "string" then
        tmp = tmp .. string.format("%q", val)
    elseif type(val) == "boolean" then
        tmp = tmp .. (val and "true" or "false")
    else
        tmp = tmp .. "\"[inserializeable datatype:" .. type(val) .. "]\""
    end

    return tmp
end

--horn functions
local function GetVhVox(vehicle)
    local vc = 0
    for i,v in pairs(GetBodyShapes(GetVehicleBody(vehicle))) do
        vc = vc + GetShapeVoxelCount(v)
    end
    return vc
end

local function GetHornType(vehicle)
    local vc = GetVhVox(vehicle)
    
    if HasTag(vehicle, "boat") then
        if vc >= 80000 then
            return "water_large"
        elseif vc >= 15000 then
            return "water_medium"
        else
            return "water_small"
        end
    else
        if vc >= 12000 then
            return "land_large"
        elseif vc >= 3100 then
            return "land_medium"
        else
            return "land_small"
        end
    end
end

--physics functions
local function IteratePhys()
    if physics.smartUpdate then
        physics.passiveBodies = {}
    end
    for i,v in pairs(global.bodies) do
        if not physics.state then
            if physics.smartUpdate and (not IsBodyActive(v)) then
                physics.passiveBodies[v] = true
            end
            SetBodyDynamic(v,false)
        else
            if not has_value(global.staticBodies, v) then
                SetBodyDynamic(v,true);
                if not physics.cancelMvmnt then
                    local com = GetBodyCenterOfMass(v)
                    local worldPoint = TransformToParentPoint(GetBodyTransform(v), com)
                    if (physics.smartUpdate) and (physics.passiveBodies[v] ~= true) then
                        ApplyBodyImpulse(v, worldPoint, imp)
                    elseif (not physics.smartUpdate) then
                        ApplyBodyImpulse(v, worldPoint, imp)
                    end
                end
            end
        end
    end
end

--misc
local function onspawnedvehicle(vehicle)
    vhdata.type[vehicle] = GetHornType(vehicle)
    vhdata.mxhp[vehicle] = GetVhVox(vehicle)
    table.insert(calarm.checkable,vehicle)
end

--init
function init()
	DefaultData(true)
	
    --global
	global = {}
	global.debug =          false
    global.tdmpdebug =      GetBool("savegame.mod.global.tdmpdebug")
    global.hasCfg =         false
    global.sentreq =        false
	global.deltaFrameTime =	0
	global.lastFrameTime = 	0
    global.shapes =         FindShapes(nil,true)
    global.bodies =         FindBodies(nil,true)
    global.staticBodies =   {}
    global.playervh =       0
    global.multiplayer =    true
    global.framelesstime =  0
    if not TDMP_LocalSteamId then 
        global.multiplayer =    false
    end
    global.ishost =         false
    if (global.multiplayer) and (TDMP_IsServer()) then
        global.ishost = true
    end

    global.fly_enabled =    GetBool("savegame.mod.global.fly.enabled")
    global.sprint_enabled = GetBool("savegame.mod.global.sprint.enabled")
    global.hands_enabled =  GetBool("savegame.mod.global.hands.enabled")
    global.phys_enabled =   GetBool("savegame.mod.global.phys.enabled")
    global.horn_enabled =   GetBool("savegame.mod.global.horn.enabled")
    global.debris_enabled = GetBool("savegame.mod.global.debris.enabled")
    global.speedmtr_enabled = GetBool("savegame.mod.global.speedmtr.enabled")
    global.calarm_enabled =  GetBool("savegame.mod.global.calarm.enabled")
    global.tctrl_enabled = GetBool("savegame.mod.global.tctrl.enabled")
    global.imoh_enabled =  GetBool("savegame.mod.global.imoh.enabled")
    global.advengine_enabled = GetBool("savegame.mod.global.advengine.enabled")
    global.sidebar =        GetBool("savegame.mod.global.sideinfo")
    global.sbkb =          GetString("savegame.mod.global.sbkb")

    for i, v in pairs(global.bodies) do
		if not IsBodyDynamic(v) then
			table.insert(global.staticBodies, v)
		end
    end

    if global.debug then
        DefaultData(false)
    end

	--fly
    --thanks to wsung and kmc for their effort on previous mods
    --https://steamcommunity.com/sharedfiles/filedetails/?id=2416253872
    --https://steamcommunity.com/sharedfiles/filedetails/?id=2419333267
    --thanks to Cr4xy for finding the magic number
    --https://steamcommunity.com/sharedfiles/filedetails/?id=2580322132
    --thanks to Fall Guy 2106 for their noclip angle/rotation code
    --https://steamcommunity.com/sharedfiles/filedetails/?id=2406955331
	fly = {}
	fly.FlyTiming = 		GetInt("savegame.mod.fly.FlyTiming")
	fly.DownKey = 			GetString("savegame.mod.fly.DownKey")
	fly.SpeedKey = 			GetString("savegame.mod.fly.SpeedKey")
	fly.NoclipKey = 		GetString("savegame.mod.fly.NoclipKey")
	fly.FreecamKey = 		GetString("savegame.mod.fly.FreecamKey")
	fly.camBaseSpeed =		GetInt("savegame.mod.fly.camBaseSpeed")
	fly.flyingMod = 		GetInt("savegame.mod.fly.flyingMod")/100
    fly.preciseMod = 		GetInt("savegame.mod.fly.preciseMod")/100
	fly.flyBaseSpeed =		GetInt("savegame.mod.fly.flyBaseSpeed")
    fly.precisionKey =      GetString("savegame.mod.fly.PreciseKey")
    fly.sandboxOnly =       GetBool("savegame.mod.fly.sandboxOnly")
    fly.healWhenDown =      GetBool("savegame.mod.fly.healWhenDown")
    fly.doDoubleTap =       GetBool("savegame.mod.fly.doDoubleTap")
    fly.flykey =            GetString("savegame.mod.fly.flykey")
    fly.cog =               GetBool("savegame.mod.fly.cog")

	fly.freecamPos = 		Vec(0, 0, 0)
	fly.input =				nil
	fly.state = 			"idle"
	fly.spaceFrameCount	=	0
	fly.SpaceCount	=		0
    fly.WasFlying =         false
    fly.PlayerToCamDiff =   0
    fly.ncAngle =           nil
    fly.lastY =             0

    if (not GetBool("level.sandbox")) and (fly.sandboxOnly) then
        global.fly_enabled = false
    end

    --hands
    hands = {}
    hands.toolgroup =       GetInt("savegame.mod.hands.toolgroup")

    if global.hands_enabled then
        RegisterTool("hands", "Hands", "MOD/vox/Void.vox", hands.toolgroup)
	    SetBool("game.tool.hands.enabled", true)
    end

    --sprint & walk
    --thanks to Cheejins for his move function
    --https://steamcommunity.com/sharedfiles/filedetails/?id=2415158477
    sprint = {}
    sprint.limit =          GetBool("savegame.mod.sprint.limit")
    sprint.time =           GetInt("savegame.mod.sprint.time")
    sprint.disableBT =      GetBool("savegame.mod.sprint.disableBT")
    sprint.sprintkey =      GetString("savegame.mod.sprint.sprintkey")
    sprint.speed =          GetInt("savegame.mod.sprint.speed")
    sprint.walkspeed =      GetInt("savegame.mod.sprint.walkspeed")/100
    sprint.walkKey =        GetString("savegame.mod.sprint.walkKey")
    
    sprint.Left =           sprint.time
    sprint.isBarUp =        false
    sprintfade =            0
    sprint.sprinting =      false
    sprint.regen =          1
    sprint.delay =          1.5
    sprint.lastRanOut =     0

    --pause physics
    physics = {}
    physics.toggleKB =      GetString("savegame.mod.physics.toggleKB")
    physics.cancelMvmnt =   GetBool("savegame.mod.physics.cancelMvmnt")
    physics.enforce =       GetBool("savegame.mod.physics.enforce")
    physics.smartUpdate =   GetBool("savegame.mod.physics.smartUpdate")
    physics.passiveBodies = {}

    physics.state =         true

    --horn
    horn = {}
    horn.hornkey =          GetString("savegame.mod.horn.hornkey")
    horn.fadespeed =        GetInt("savegame.mod.horn.fadespeed")
    
    horn.ishorn =           false
    horn.type =             nil
    horn.fadedir =          true --false = in, true = out
    horn.fade =             0
    hornsounds = {}
    hornsounds.land_small = LoadLoop("MOD/sound/land_small.ogg", 10)
    hornsounds.land_medium = LoadLoop("MOD/sound/land_medium.ogg", 25)
    hornsounds.land_large = LoadLoop("MOD/sound/land_large.ogg", 50)
    hornsounds.water_small = LoadLoop("MOD/sound/water_small.ogg", 10)
    hornsounds.water_medium = LoadLoop("MOD/sound/water_medium.ogg", 50)
    hornsounds.water_large = LoadLoop("MOD/sound/water_large.ogg", 125)
    vhdata = {}
    vhdata.vehicles = FindVehicles(nil,true)
    vhdata.type = {}
    for i,v in pairs(vhdata.vehicles) do
        vhdata.type[v] = GetHornType(v)
    end

    horn.horningVH = {} --List vehicles horning, for TDMP compat

    --debris
    debris = {}
    debris.bottom = GetBool("savegame.mod.debris.bottom")
    debris.top = GetBool("savegame.mod.debris.top")
    debris.persist = {"tool", "tools", "section", "valuable", "interact", "debrisIgnore"}
    debris.terminalScreen = FindScreen("terminal", true)
    debris.terminalBody = GetShapeBody(GetScreenShape(debris.terminalScreen))
    debris.count = 0
    debris.tick = 0
    debris.lightmode = GetBool("savegame.mod.debris.lightmode")

    --speedometer
    speedmtr = {}
    speedmtr.kph = GetBool("savegame.mod.speedmtr.kph")
    speedmtr.lim = GetInt("savegame.mod.speedmtr.lim")

    --car alarm
    calarm = {}
    calarm.limit = GetInt("savegame.mod.calarm.limit")/100 --0.8
    calarm.doEngineCycle = GetBool("savegame.mod.calarm.doEngineCycle")
    calarm.wateralarm = GetBool("savegame.mod.calarm.wateralarm")
    calarm.checkable = vhdata.vehicles
    calarm.repeatalarms = true
    calarm.init = {}
    calarm.state = {}

    vhdata.mxhp = {}
    for i,v in pairs(vhdata.vehicles) do
        vhdata.mxhp[v] = GetVhVox(v)
    end
    
    --time control
    tctrl = {}
    tctrl.incrmnt = GetInt("savegame.mod.tctrl.incrmnt")/100
    tctrl.up = GetString("savegame.mod.tctrl.up")
    tctrl.down = GetString("savegame.mod.tctrl.down")
    tctrl.toggle = GetString("savegame.mod.tctrl.toggle")
    tctrl.fade = GetBool("savegame.mod.tctrl.fade")
    tctrl.immune = GetBool("savegame.mod.tctrl.immune")

    tctrl.cur = 1
    tctrl.active = false
    tctrl.transamnt =     1
    tctrl.instrans =      false

    --I'm outta here
    imoh = {}
    imoh.vel = GetInt("savegame.mod.imoh.vel")
    imoh.key = GetString("savegame.mod.imoh.key")

    --Recommended overrides for TDMP
    if global.multiplayer and global.ishost and GetBool("savegame.mod.global.tdmpoverides") then
        --pause physics
        physics.enforce =       true
        physics.smartUpdate =   false

        --time control
        tctrl.fade = false
    end

    --Register TDMP stuff
    if global.multiplayer then
        TDMP_RegisterEvent("VariableSync", function(jsonData, sender)
            local data = json.decode(jsonData)
            local varid = data[1]
            local varsub = data[2]
            local value = data[3]
            if global.tdmpdebug then
                DebugPrint("[AU-TDMP] Set root:"..varid..",sub:"..varsub.." to:"..tostring(value))
            end
            if varid == 1 then
                physics[varsub] = value
                IteratePhys()
            elseif varid == 2 then
                tctrl[varsub] = value
            end
            if not TDMP_IsServer() then return end
            TDMP_ServerStartEvent("VariableSync", {
                Receiver = TDMP.Enums.Receiver.ClientsOnly,
                Reliable = true,
                DontPack = true,
                Data = jsonData
            })
        end)
        TDMP_RegisterEvent("ConfigSync", function(jsonData, sender)
            local tosync = {
                {"fly_enabled","sprint_enabled","hands_enabled","phys_enabled","horn_enabled","debris_enabled","speedmtr_enabled","calarm_enabled","tctrl_enabled","imoh_enabled"}, --global
                {"fade"}, --tctrl
                {"cancelMvmnt", "enforce", "smartUpdate"} --physics
            }
            if not global.ishost then 
                if global.tdmpdebug then
                    DebugPrint("[AU-TDMP] Recieved config sync request form server...")
                end
                if not global.hasCfg then
                    local data = json.decode(jsonData)
                    for i,v in pairs(tosync) do
                        for m,n in pairs(v) do
                            if i == 1 then
                                global[n] = data[i][m]
                            elseif i == 2 then
                                tctrl[n] = data[i][m]
                            elseif i == 3 then
                                physics[n] = data[i][m]
                            end
                            if global.tdmpdebug then
                                DebugPrint("[AU-TDMP] Sycing config - "..tostring(data[i][m]).." - ("..m.."/"..#v..")("..i.."/"..#tosync..")...")
                            end
                        end
                    end
                    if global.tdmpdebug then
                        DebugPrint("[AU-TDMP] Synced config with server!")
                    end
                    global.hasCfg = true
                end
            else
                if global.tdmpdebug then
                    DebugPrint("[AU-TDMP] Someone requested config...")
                end
                local send = tosync
                for i,v in pairs(tosync) do
                    for m,n in pairs(v) do
                        if i == 1 then
                            send[i][m] = global[n]
                        elseif i == 2 then
                            send[i][m] = tctrl[n]
                        elseif i == 3 then
                            send[i][m] = physics[n]
                        end
                    end
                end
                if global.tdmpdebug then
                    DebugPrint("[AU-TDMP] Sending config to client "..sender)
                end
                TDMP_ServerStartEvent("ConfigSync", {
                    Receiver = TDMP.Enums.Receiver.ClientsOnly,
                    Reliable = true,
                    DontPack = false,
                    Data = send
                })
            end
        end)
        TDMP_RegisterEvent("HornSync", function(jsonData, sender)
            local data = json.decode(jsonData)
            local vehicle = data[1]
            local state = data[2]
            local player
            if global.ishost then
                player = sender
            else
                player = data[3]
            end
            if global.tdmpdebug then
                DebugPrint("[AU-TDMP] Player "..tostring(player).." set horn for "..tostring(vehicle).." to "..tostring(state))
            end
            if horn.horningVH[player] == nil then
                horn.horningVH[player] = {}
            end
            horn.horningVH[player]["vehicle"] = vehicle
            horn.horningVH[player]["state"] = state
            --if global.tdmpdebug then
            --    DebugPrint(serializeTable(horn.horningVH))
            --end
            if not TDMP_IsServer() then return end
            table.insert(data, sender)
            TDMP_ServerStartEvent("HornSync", {
                Receiver = TDMP.Enums.Receiver.ClientsOnly,
                Reliable = true,
                DontPack = false,
                Data = data
            })
        end)
    end
    if global.tdmpdebug then
        DebugPrint("[AU-TDMP] Finished init...")
    end

    --engine stuff
    advengine = {}
    advengine.startreqkey = false --GetBool("savegame.mod.advengine.startreqkey")
    advengine.doautopark = GetBool("savegame.mod.advengine.doautopark")

    advengine.vehicles = {}
    advengine.isincar = false
    advengine.beepbeep = LoadSound("MOD/sound/carlock.ogg", 0)
    if global.advengine_enabled then
        RegisterTool("carkeys", "Car Keys", "MOD/vox/Void.vox", 6)
        SetBool("game.tool.carkeys.enabled", true)
    end
end

--fly functions
local function holdPlayerStill()
	local curvec = GetPlayerVelocity()
	local modifiedvec = {curvec[1],curvec[2],curvec[3]}
	local factor = 200/100
	local reduxfactor = 50/100
	if curvec[2] < 0 then
		modifiedvec[2] = 0.16663932800293
	else
		modifiedvec[2] = modifiedvec[2]*reduxfactor
	end
	modifiedvec[1] = modifiedvec[1]*reduxfactor
	modifiedvec[3] = modifiedvec[3]*reduxfactor
	SetPlayerVelocity(Vec(modifiedvec[1], 0.16663932800293, modifiedvec[3]))
end

local function syncCamAndPlayer(init, tools)
    if (init) then
        fly.PlayerToCamDiff = GetPlayerCameraTransform().pos[2]-GetPlayerTransform().pos[2]
        if tools then
            fly.freecamPos = GetPlayerTransform().pos
        else
            fly.freecamPos = GetPlayerCameraTransform().pos
        end
    else
        SetPlayerTransform(Transform(Vec(fly.freecamPos[1], fly.freecamPos[2]-fly.PlayerToCamDiff, fly.freecamPos[3]), GetCameraTransform(true).rot), true)
    end
end

local function freeCam(tools)
    local cameraFront = Vec(0, 0, 0)
    local cameraUp = Vec(0, 1, 0)
    
    local mult = 1
    if (((tctrl.active) and (tctrl.cur<1)) or ((tctrl.instrans) and (tctrl.transamnt<1))) and tctrl.immune then
        if tctrl.instrans then
            mult = 1/tctrl.transamnt
        else
            mult = 1/tctrl.cur
        end
    end

    local camSpeed = fly.input.camSpeed * global.deltaFrameTime * mult
    local cameraFront = getCameraFront()

    if (InputDown("up")) then
        fly.freecamPos = VecAdd(fly.freecamPos, VecScale(cameraFront, camSpeed))
    end
    if (InputDown("down")) then
        fly.freecamPos = VecSub(fly.freecamPos, VecScale(cameraFront, camSpeed))
    end
    if (InputDown("left")) then
        fly.freecamPos = VecSub(fly.freecamPos, VecScale(VecNormalize(VecCross(cameraFront, cameraUp)), camSpeed))
    end
    if (InputDown("right")) then
        fly.freecamPos = VecAdd(fly.freecamPos, VecScale(VecNormalize(VecCross(cameraFront, cameraUp)), camSpeed))
    end
    if (InputDown("space")) then
        fly.freecamPos = VecAdd(fly.freecamPos, Vec(0, camSpeed, 0))
    end
    if (InputDown(fly.DownKey)) then
        fly.freecamPos = VecSub(fly.freecamPos, Vec(0, camSpeed, 0))
    end

    local lookat = QuatLookAt(fly.freecamPos, VecAdd(fly.freecamPos, cameraFront))

    if tools then
        local dx, dy = InputValue("mousedx"), InputValue("mousedy")
        local sensitivity = GetInt("options.input.sensitivity")/100

        fly.ncAngle[1] = math.min(math.max(fly.ncAngle[1] - dy / 10 * sensitivity,-80), 80)
        fly.ncAngle[2] = fly.ncAngle[2] - dx / 10 * sensitivity
        lookat = QuatEuler(fly.ncAngle[1], fly.ncAngle[2], 0)
        SetPlayerTransform(Transform(fly.freecamPos, lookat), true)
    else
        SetCameraTransform(Transform(fly.freecamPos, lookat))
    end
end

local function flyfunc()
    if (fly.input.moving) then
        local playerTransform = GetPlayerTransform()
        local parent = TransformToParentPoint(playerTransform, fly.input.direction)
        local reverse = VecNormalize(VecSub(parent, playerTransform.pos))

        local mult = 1
        if (((tctrl.active) and (tctrl.cur<1)) or ((tctrl.instrans) and (tctrl.transamnt<1))) and tctrl.immune then
            if tctrl.instrans then
                mult = 1/tctrl.transamnt
            else
                mult = 1/tctrl.cur
            end
        end

        local velocity = VecScale(reverse, fly.input.flyingVel*mult)
        local tempvely = velocity[2]
        if tempvely == 0 then
            tempvely = 0.16663932800293
        end
        SetPlayerVelocity(Vec(velocity[1], tempvely, velocity[3]))
    else
        holdPlayerStill()
    end
end

local function flyCheck()
    local input = {}

    input.direction = Vec(0, 0, 0)

    input.moving = true
	input.flyingVel = fly.flyBaseSpeed 
    input.camSpeed = fly.camBaseSpeed

    if (InputDown("up")) then
        input.direction = VecAdd(input.direction, Vec(0, 0, -1))
    end

    if (InputDown("left")) then
        input.direction = VecAdd(input.direction, Vec(-1, 0, 0))
    end

    if (InputDown("down")) then
        input.direction = VecAdd(input.direction, Vec(0, 0, 1))
    end

    if (InputDown("right")) then
        input.direction = VecAdd(input.direction, Vec(1, 0, 0))
    end

    if (InputDown("space")) then
        input.direction = VecAdd(input.direction, Vec(0, 1, 0))
    end

    if (InputDown(fly.DownKey)) then
        input.direction = VecAdd(input.direction, Vec(0, -1, 0))
    end

    if InputDown(fly.SpeedKey) then
        input.flyingVel = fly.flyBaseSpeed * fly.flyingMod
        input.camSpeed = fly.camBaseSpeed * fly.flyingMod
    end
    if InputDown(fly.precisionKey) then
        input.flyingVel = fly.flyBaseSpeed * fly.preciseMod
        input.camSpeed = fly.camBaseSpeed * fly.preciseMod
    end

    if (input.direction[1] == 0 and input.direction[2] == 0 and input.direction[3] == 0) then
        input.moving = false
    end

    return input
end

--sprint and walk functions
local function MovePlayer(speed, dir, fast)
    local curspeed = math.sqrt(GetPlayerVelocity()[1]^2+GetPlayerVelocity()[3]^2)
    local mvDir = VecScale(dir, speed)

    local camTr = GetPlayerCameraTransform()
    local camFwd = TransformToParentPoint(camTr, Vec(0,0,-1))
    camTr.pos[2] = 0
    camFwd[2] = 0

    local camSubFlatY = VecNormalize(VecSub(camTr.pos, camFwd))
    camTr.rot = QuatLookAt(GetPlayerTransform(), camSubFlatY)

    local sVel = GetPlayerVelocity()
    sVel = VecSub(camTr.pos, TransformToParentPoint(camTr, mvDir))
    sVel[2] = GetPlayerVelocity()[2]
    local nextspeed = math.sqrt(sVel[1]^2+sVel[3]^2)
    if (fast) and (curspeed>=nextspeed) then return end
    SetPlayerVelocity(sVel)
    -- printVec(GetPlayerVelocity())
end

--car alarm functions
local function SetBodyEmissiveScale(handle, scale)
    for i,v in pairs(GetBodyShapes(handle)) do
        SetShapeEmissiveScale(v, scale)
    end
end

--debris functions
local function CheckTags(tags, handle)
    for i,v in pairs(tags) do
        if (HasTag(handle, v)) then
            return true
        end
    end
    return false
end

--tick
function tick(dt)
    --global stuff
    local FrameTime = GetTime()
    global.deltaFrameTime = FrameTime - global.lastFrameTime
    global.lastFrameTime = FrameTime
    debris.tick = debris.tick + 1
    global.framelesstime = global.framelesstime + global.deltaFrameTime

	--fly stuff
    fly.input = flyCheck()

	if InputPressed(fly.flykey) and (global.fly_enabled) then
        if fly.doDoubleTap then
            fly.SpaceCount =  fly.SpaceCount + 1
        else
            fly.SpaceCount =  2
        end
    end
    if InputPressed(fly.NoclipKey) and (fly.state ~= "freecam") and (global.playervh == 0) and (global.fly_enabled) then
        if (fly.state == "idle") or (fly.state == "flying") then
            fly.WasFlying = (fly.state == "flying")
            syncCamAndPlayer(true, true)
            fly.ncAngle = Vec(toEuler(GetCameraTransform().rot))
            fly.state = "noclip"
        else
            if fly.WasFlying then
                fly.state = "flying"
            else
                fly.state = "idle"
            end
        end
    end
    if InputPressed(fly.FreecamKey) and (fly.state ~= "noclip") and (global.playervh == 0) and (global.fly_enabled) then
        if (fly.state == "idle") or (fly.state == "flying") then
            fly.WasFlying = (fly.state == "flying")
            syncCamAndPlayer(true, false)
            fly.state = "freecam"
        else
            if fly.WasFlying then
                fly.state = "flying"
            else
                fly.state = "idle"
            end
        end
    end

    if (fly.state == "noclip") or (fly.state == "freecam") then
	    freeCam((fly.state == "noclip"))
        if (fly.state == "noclip") then
        
        else
            holdPlayerStill()
            SetPlayerHealth(1)
        end
	end

    if (fly.healWhenDown or fly.cog) and (InputDown(fly.DownKey)) and (not InputDown("space")) and (fly.state == "flying") then
        if fly.healWhenDown then
            SetPlayerHealth(1)
        end
        if fly.cog then
            local hit, dist, normal, shape = QueryRaycast(GetPlayerTransform().pos, Vec(0, -1, 0), 1)
            if(dist < 1) then
                fly.state = "idle"
                SetPlayerHealth(1)
            end
        end
    end

    --sprint
    if InputDown(sprint.sprintkey) and (global.playervh == 0) and (fly.state == "idle") and ((sprint.Left > 0) or (not sprint.limit)) and (global.sprint_enabled) then
        sprint.sprinting = true
        sprint.lastRanOut = GetTime()
        if fly.input.moving then
            MovePlayer(sprint.speed, fly.input.direction, true)
            sprint.Left = sprint.Left - global.deltaFrameTime
        end
    else
        sprint.sprinting = false
        if (sprint.Left < sprint.time) and (GetTime() > (sprint.lastRanOut + sprint.delay)) then
            sprint.Left = sprint.Left + sprint.regen * global.deltaFrameTime
        elseif sprint.Left > sprint.time then
            sprint.Left = sprint.time
        end
    end
    if InputDown(sprint.walkKey) and (global.playervh == 0) and (fly.state == "idle") and (global.sprint_enabled) then
        if fly.input.moving then
            MovePlayer(sprint.walkspeed, fly.input.direction, false)
        end
    end

    --pause physics
    if InputPressed(physics.toggleKB) and (global.phys_enabled) then
        physics.state = not physics.state
        if global.multiplayer then
            if global.tdmpdebug then
                DebugPrint("[AU-TDMP] Pushing physics state")
            end
            TDMP_ClientStartEvent("VariableSync", {
                Reliable = true,
    
                DontPack = false,
                Data = {1, "state", physics.state}
            })
        else
            IteratePhys()
        end
    end

    if (not physics.state) and (physics.enforce) then
        for i,v in pairs(global.bodies) do
            SetBodyDynamic(v,false)
        end
    end

    --horn
    if not global.multiplayer then
        if InputDown(horn.hornkey) and (global.playervh ~= 0) and (global.horn_enabled) then
            if not horn.ishorn then
                if vhdata.type[global.playervh] == nil then
                    vhdata.type[global.playervh] = GetHornType(global.playervh)
                end
                horn.type = vhdata.type[global.playervh]
                horn.fadedir = false --false = in, true = out
                horn.fade = 0
                horn.ishorn = true
            end
        elseif (horn.ishorn) and (not horn.fadedir) then
            horn.fadedir = true --false = in, true = out
        end
    else
        if InputDown(horn.hornkey) and (global.playervh ~= 0) and (global.horn_enabled) then
            if (not horn.ishorn) then
                TDMP_ClientStartEvent("HornSync", {
                    Reliable = true,
                    DontPack = false,
                    Data = {global.playervh, true}
                })
                if global.tdmpdebug then
                    DebugPrint("[AU-TDMP] Pushing horn state")
                end
            end
            horn.ishorn = true
        elseif (horn.ishorn) then
            TDMP_ClientStartEvent("HornSync", {
                Reliable = true,
                DontPack = false,
                Data = {global.playervh, false}
            })
            if global.tdmpdebug then
                DebugPrint("[AU-TDMP] Pushing horn state")
            end
            horn.ishorn = false
        end
    end
    

    if not global.multiplayer then
        if (not horn.fadedir) and (horn.fade < 1) then
            horn.fade = horn.fade + global.deltaFrameTime*horn.fadespeed
        elseif (horn.fadedir) and (horn.fade > 0) then
            horn.fade = horn.fade - global.deltaFrameTime*horn.fadespeed
        end
        if horn.fade>1 then horn.fade = 1 end
        if horn.fade<0 then horn.fade = 0 end
        if horn.fade==0 and horn.ishorn then horn.ishorn = false end

        if horn.ishorn then
            local vhbod = GetVehicleBody(global.playervh)
            local vhpos = TransformToParentPoint(GetBodyTransform(vhbod), GetBodyCenterOfMass(vhbod))
            PlayLoop(hornsounds[horn.type], vhpos, horn.fade)
        end
    else
        for i,v in pairs(horn.horningVH) do
            if v["state"] then
                local vhbod = GetVehicleBody(v["vehicle"])
                local vhpos = TransformToParentPoint(GetBodyTransform(vhbod), GetBodyCenterOfMass(vhbod))
                if vhdata.type[v["vehicle"]] == nil then
                    vhdata.type[v["vehicle"]] = GetHornType(v["vehicle"])
                end
                PlayLoop(hornsounds[vhdata.type[v["vehicle"]]], vhpos, 1)
            end
        end
    end

    --debris
    if (global.debris_enabled and physics.state) and ((debris.lightmode and debris.tick >= 1000) or (not debris.lightmode)) then
        for i,v in pairs(global.bodies) do
            local center = TransformToParentPoint(GetBodyTransform(v), GetBodyCenterOfMass(v)) 
            if ((center[2] <= -19) and (debris.bottom)) or ((center[2] >= 100000) and (debris.top)) then
                if (CheckTags(debris.persist, v) == false) and (v ~= debris.terminalBody) and (v ~= GetToolBody()) then
                    Delete(v)
                    debris.count = debris.count + 1
                end
            end
        end
        debris.tick = 0
    end

    --car alarm
    if global.calarm_enabled then
        if (global.playervh ~= 0) then --If a player enters a vehicle the alarm stops/resets
            local ind = find_value(calarm.checkable, global.playervh) 
            if (ind ~= false) then
                if calarm.repeatalarms then
                    vhdata.mxhp[global.playervh] = GetVhVox(global.playervh)
                else
                    table.remove(calarm.checkable, ind)
                end
                calarm.state[global.playervh] = false
            end
        end
    
        local nt = {}
        for i,v in pairs(calarm.checkable) do
            if advengine.vehicles[v] == nil then
                advengine.vehicles[v] = {false, advengine.doautopark}
            end
            local hp = GetVhVox(v) / vhdata.mxhp[v]

            local body = GetVehicleBody(v)
            local center = TransformToParentPoint(GetBodyTransform(body), GetBodyCenterOfMass(body))
            local wet, d = IsPointInWater(center)
            if HasTag(v, "boat") then
                wet = false
            end

            if hp < calarm.limit then
                
            elseif (((wet) and (calarm.wateralarm)) or (hp < 1)) and (not advengine.vehicles[v][1]) then
                table.insert(nt, v)
                if global.debug then
                    DrawBodyOutline(body, 1, 1, 1, 1)
                end
                if calarm.state[v] ~= true then
                    calarm.state[v] = true
                    calarm.init[v] = global.framelesstime
                end
                local loudness = math.sin(10*(global.framelesstime - calarm.init[v]))/2+0.5
                local vhbod = body
                local vhpos = center
                if global.debug then
                    DebugCross(vhpos, 1, 1, 1, 1)
                end
                PlayLoop(hornsounds[vhdata.type[v]], vhpos, loudness)
                SetBodyEmissiveScale(vhbod, loudness)
                if (not calarm.doEngineCycle) or (loudness >= 0.5) then
                    DriveVehicle(v, 0.000001, 0, false)
                end
            else
                table.insert(nt, v)
            end
        end
        calarm.checkable = nt
    end

    --time control
    if InputDown(tctrl.up) and (tctrl.cur < 1) then
        tctrl.cur = tctrl.cur + tctrl.incrmnt
        if tctrl.cur > 1 then tctrl.cur = 1 end
        
        if global.multiplayer then
            if global.tdmpdebug then
                DebugPrint("[AU-TDMP] Pushing timecontrol cur")
            end
            TDMP_ClientStartEvent("VariableSync", {
                Reliable = true,
                DontPack = false,
                Data = {2, "cur", tctrl.cur}
            })
        end
    elseif InputDown(tctrl.down) and (tctrl.cur > 0.1) then
        tctrl.cur = tctrl.cur - tctrl.incrmnt
        if tctrl.cur < 0.1 then tctrl.cur = 0.1 end
        
        if global.multiplayer then
            if global.tdmpdebug then
                DebugPrint("[AU-TDMP] Pushing timecontrol cur")
            end
            TDMP_ClientStartEvent("VariableSync", {
                Reliable = true,
                DontPack = false,
                Data = {2, "cur", tctrl.cur}
            })
        end
    end

    if InputPressed(tctrl.toggle) then
        tctrl.active = not tctrl.active
        if global.multiplayer then
            if global.tdmpdebug then
                DebugPrint("[AU-TDMP] Pushing timecontrol state")
            end
            TDMP_ClientStartEvent("VariableSync", {
                Reliable = true,
                DontPack = false,
                Data = {2, "active", tctrl.active}
            })
        end
        if tctrl.fade then
            tctrl.instrans = true
        end
    end

    if (tctrl.instrans) then
        if tctrl.active then
            if (tctrl.transamnt > tctrl.cur) then
                tctrl.transamnt = tctrl.transamnt - global.deltaFrameTime
            else
                tctrl.instrans = false
                tctrl.transamnt = tctrl.cur
            end
        else
            if (tctrl.transamnt < 1) then
                tctrl.transamnt = tctrl.transamnt + global.deltaFrameTime
            else
                tctrl.instrans = false
                tctrl.transamnt = 1
            end
        end
        SetTimeScale(tctrl.transamnt)
    elseif (tctrl.active) then
        SetTimeScale(tctrl.cur)
    end

    if (global.imoh_enabled) and (InputPressed(imoh.key)) then
		if GetPlayerVehicle() == 0 then
            local cur = GetPlayerVelocity()
			SetPlayerVelocity(Vec(cur[1] * imoh.vel, cur[2] * imoh.vel, cur[3] * imoh.vel))
		else
            local cur = GetBodyVelocity(GetVehicleBody(GetPlayerVehicle()))
			SetBodyVelocity(GetVehicleBody(GetPlayerVehicle()), Vec(cur[1] * imoh.vel, cur[2] * imoh.vel, cur[3] * imoh.vel))
		end
	end

    --Sidebar toggle
    if InputPressed(global.sbkb) then
        global.sidebar = not global.sidebar
    end

    if global.multiplayer then
        TDMP_Hook_Queue()
    end

    --engine stuff
    if global.advengine_enabled then
        if global.playervh ~= 0 and (not advengine.isincar) then
            --check if car is indexed
            if advengine.vehicles[global.playervh] == nil then
                advengine.vehicles[global.playervh] = {(not advengine.startreqkey), advengine.doautopark}
            else
                advengine.vehicles[global.playervh][1] = not advengine.startreqkey
            end
            advengine.isincar = true
        end
        if global.playervh == 0 and advengine.isincar then
            advengine.isincar = false
        end
    
        for vehicleid,enginestate in pairs(advengine.vehicles) do 
            if GetVehicleTransform(vehicleid) == nil then
                table.remove(advengine.vehicles, vehicleid)
            elseif enginestate[1] then
                if (vehicleid ~= global.playervh) and (GetVehicleHealth(vehicleid) > 0) then
                    DriveVehicle(vehicleid, 0.000001, 0, enginestate[2])
                end
            end
        end
    end
end

--tdmp
if TDMP_LocalSteamId then
    Hook_AddListener("ConnectedToServer", "conf", function() 
        --Config sync
        if (not global.ishost) and (not global.sentreq) then --request config sync
            TDMP_ClientStartEvent("ConfigSync", {
                Reliable = true,
                DontPack = true,
                Data = "give me config >:("
            })
            if global.tdmpdebug then
                DebugPrint("[AU-TDMP] Requesting config...")
            end
            global.sentreq = true
        end
    end)
end

--update
function update(dt)
	--global stuff
    if global.debug then
		DebugWatch("global.deltaFrameTime", global.deltaFrameTime)
		DebugWatch("global.lastFrameTime", global.lastFrameTime)
        DebugWatch("framelesstime", global.framelesstime)
	end

    global.shapes = FindShapes(nil,true)
    global.bodies = FindBodies(nil,true)
    global.playervh = GetPlayerVehicle()

    local vhquerry = FindVehicles(nil,true)
    if #vhdata.vehicles < #vhquerry then
        for index,vh in pairs(vhquerry) do
            if not has_value(vhdata.vehicles, vh) then
                onspawnedvehicle(vh)
            end
        end
        vhdata.vehicles = vhquerry
    end
    
    --fly stuff
	if (fly.state == "flying") then
        flyfunc()
    end

    if (fly.SpaceCount >= 1) then
        fly.spaceFrameCount = fly.spaceFrameCount + 1
    end

    if (fly.spaceFrameCount > fly.FlyTiming) then
        fly.spaceFrameCount = 0
        fly.SpaceCount = 0
    end

    if (fly.SpaceCount >= 2) and (global.playervh == 0) then
		if (fly.state == "flying") then
			fly.state = "idle"
			holdPlayerStill()
		elseif (fly.state == "noclip") or (fly.state == "freecam") then
			fly.state = "idle"
		else
			fly.state = "flying"
		end
		fly.spaceFrameCount = 0
        fly.SpaceCount = 0
    end

    if fly.state == "noclip" then        --detect going outside map in noclip and remedy it
        local precision = 4
        local currentpos = GetPlayerTransform().pos
        local dir = {"x","y","z"}
        for i=1,3 do 
            if (round(fly.freecamPos[i], precision) ~= round(currentpos[i], precision)) then
                --DebugPrint("NC discrepancy of ".. round(fly.freecamPos[i], precision)-round(currentpos[i], precision) .. " in " .. dir[i])
                fly.freecamPos[i] = currentpos[i]
                if (i==2) and (fly.freecamPos[i]<-19) then --so the player can not go below the level
                    fly.freecamPos[i] = -19
                end
            end
        end
    end

	if global.debug then
		DebugWatch("fly.freecamPos", fly.freecamPos)
		DebugWatch("fly.input.moving", fly.input.moving)
		DebugWatch("fly.input", fly.input)
		DebugWatch("fly.state", fly.state)
		DebugWatch("fly.spaceFrameCount", fly.spaceFrameCount)
		DebugWatch("fly.SpaceCount", fly.SpaceCount)
        DebugWatch("fly.input.flyingVel", fly.input.flyingVel)
	end

    --sprint
    if global.debug then
		DebugWatch("sprint.isBarUp", sprint.isBarUp)
        DebugWatch("ishUp", (GetFloat("game.player.health") < 1))
        DebugWatch("sprint.sprintFade", sprintfade)
        DebugWatch("sprint.Left", sprint.Left)
        DebugWatch("target", sprint.lastRanOut + sprint.delay)
        DebugWatch("cur", GetTime())
        DebugWatch("shouldregen", (GetTime() > (sprint.lastRanOut + sprint.delay)))
	end

    --physics
    if global.debug then
        DebugWatch("physics.state", physics.state)
    end

    --horn
    if global.debug then
        DebugWatch("horn.ishorn",horn.ishorn)
        DebugWatch("horn.type",horn.type)
        DebugWatch("horn.fadedir",horn.fadedir)
        DebugWatch("horn.fade",horn.fade)
        local vc = 0
        for i,v in pairs(GetBodyShapes(GetVehicleBody(global.playervh))) do
            vc = vc + GetShapeVoxelCount(v)
        end
        DebugWatch("vehicle voxcount", vc)
    end

    --time control
    if global.debug then
        DebugWatch("tctrl.active",tctrl.active)
        DebugWatch("tctrl.cur",tctrl.cur)
        DebugWatch("tctrl.transamnt", tctrl.transamnt)
    end
end

--draw
function draw()
	--Sprint & walk stuff
    if sprint.disableBT and (global.sprint_enabled) then
        SetBool("game.tool.steroid.enabled", false)
    end

    if (sprint.sprinting or (sprint.Left ~= sprint.time)) and (sprint.limit) and (global.sprint_enabled) and (global.playervh == 0) then
        local offset = 0
        local cur = sprint.Left/sprint.time
        local ishUp = (GetFloat("game.player.health") < 1)
        if sprintfade ~= 0 then
            offset = 35
        end
        if (ishUp) and (not sprint.isBarUp) then
            SetValue("sprintfade", 1, "easeout", 0.5)
        elseif (not ishUp) and (sprint.isBarUp) then
            SetValue("sprintfade", 0, "easein", 0.5)
        end
        sprint.isBarUp = ishUp

        UiPush() --Taken from hud.lua at line 1354
            UiTranslate(UiWidth() - 144, UiHeight() - (44 + offset*sprintfade))

            UiColor(0,0,0, 0.5)
            UiPush()
                UiColor(1,1,1)
                UiFont("bold.ttf", 24)
                UiTranslate(0, 22)
                if cur < 0.1 then
                    if math.mod(GetTime(), 1.0) < 0.5 then
                        UiColor(1, 0, 0,  1.0)
                    else
                        UiColor(1, 0, 0,  0.1)
                    end
                elseif cur < 0.5 then
                    UiColor(1, 0, 0)
                end
                UiAlign("right")
                UiText("SPRINT")
            UiPop()

            UiTranslate(10, 4)
            local w = 110
            local h = 20
            UiPush()
                UiAlign("left top")
                UiColor(0, 0, 0, 0.5)
                UiImageBox("ui/common/box-solid-10.png", w, h, 6, 6)
                if cur > 0 then
                    UiTranslate(2, 2)
                    w = (w-4)*cur
                    if w < 12 then w = 12 end
                    h = h-4
                    UiColor(1,cur*2,cur,1)
                    UiImageBox("ui/common/box-solid-6.png", w, h, 6, 6)
                end
            UiPop()

        UiPop()
    end

    --horn in vehicle controls
    if (global.playervh ~= 0) and (global.horn_enabled) then
        UiPush()
            local offset = 0
            if (GetFloat("game.player.health") < 1) then
                offset = 50
            end
            --hud.lua line 700
            local vehicle = global.playervh

            local interact  = string.upper(GetString("game.input.interact.icon"))		if interact  == "" then	interact  = string.upper(GetString("game.input.interact"))		end
            local handbrake = string.upper(GetString("game.input.handbrake.icon"))		if handbrake == "" then	handbrake = string.upper(GetString("game.input.handbrake"))		end
            local action	= string.upper(GetString("game.input.vehicle_action.icon"))	if action	 == "" then	action	  = string.upper(GetString("game.input.vehicle_action"))end
            local lmb		= string.upper(GetString("game.input.vehicle_raise.icon"))	if lmb		 == "" then	lmb		  = string.upper(GetString("game.input.vehicle_raise"))	end
            local rmb		= string.upper(GetString("game.input.vehicle_lower.icon"))	if rmb		 == "" then	rmb		  = string.upper(GetString("game.input.vehicle_lower"))	end
            local fwd		= string.upper(GetString("game.input.up.icon"))				if fwd		 == "" then	fwd		  = string.upper(GetString("game.input.up"))			end
            local bwd		= string.upper(GetString("game.input.down.icon"))			if bwd		 == "" then	bwd		  = string.upper(GetString("game.input.down"))			end
            local left		= string.upper(GetString("game.input.left.icon"))			if left		 == "" then	left	  = string.upper(GetString("game.input.left"))			end
            local right		= string.upper(GetString("game.input.right.icon"))			if right	 == "" then	right	  = string.upper(GetString("game.input.right"))			end
            if left == right then
                right = ""
            end
            local lmbrmb = lmb.." "..rmb
            local info = {}
            if HasTag(vehicle, "cranetop") then
                info[#info+1] = {left.." "..right, "Turn"}
                info[#info+1] = {fwd.." "..bwd, "Extend"}
                info[#info+1] = {lmbrmb, "Raise/lower"}
                info[#info+1] = {action, "Hook"}
            else
                info[#info+1] = {fwd, "Accelerate"}
                info[#info+1] = {bwd, "Reverse"}
                info[#info+1] = {left, "Steer"}
                if HasTag(vehicle, "cranebottom") then
                    info[#info+1] = {lmbrmb, "Support legs"}
                    info[#info+1] = {handbrake, "Brake"}
                elseif HasTag(vehicle, "crane") then
                    info[#info+1] = {lmbrmb, "Arm"}
                    info[#info+1] = {action, "Hook"}
                    info[#info+1] = {handbrake, "Brake"}
                elseif HasTag(vehicle, "dumptruck") then
                    info[#info+1] = {lmbrmb, "Bed"}
                    info[#info+1] = {handbrake, "Brake"}
                elseif HasTag(vehicle, "frontloader") then
                    info[#info+1] = {lmbrmb, "Shovel"}
                    info[#info+1] = {handbrake, "Brake"}
                elseif HasTag(vehicle, "skylift") then
                    info[#info+1] = {lmbrmb, "Lift"}
                    info[#info+1] = {handbrake, "Brake"}
                elseif HasTag(vehicle, "forklift") then
                    info[#info+1] = {lmbrmb, "Fork"}
                    info[#info+1] = {handbrake, "Brake"}
                elseif HasTag(vehicle, "tank") then
                    info[#info+1] = {lmb, "Fire"}
                    info[#info+1] = {handbrake, "Brake"}
                elseif HasTag(vehicle, "boat") then
                else
                    info[#info+1] = {handbrake, "Handbrake"}
                end
            end
            info[#info+1] = {interact, "Exit vehicle"}

            local kw = 0
            local vw = 0
            for i=1, #info do
                UiFont("bold.ttf", 22)
                local w0 = 50
                if string.sub(info[i][1], 1, 4) ~= "RAW:" then
                    w0 = UiGetTextSize(info[i][1])
                end
                UiFont("regular.ttf", 22)
                local w1 = UiGetTextSize(info[i][2])
                kw = math.max(kw,w0)
                vw = math.max(vw,w1)
            end

            local w = kw + vw + 50
            local h = #info*22 + 30
            local ah = 42
            UiAlign("top left")
            UiTranslate(UiWidth()-w-20, UiHeight()-ah-20-h-10-offset)
            UiColor(0+(horn.fade),0+(horn.fade),0+(horn.fade),0.5)
            UiImageBox("ui/common/box-solid-6.png", w, ah, 6, 6)
            UiTranslate(kw + 20, 28)
            UiColor(1-(horn.fade),1-(horn.fade),1-(horn.fade))
            UiFont("bold.ttf", 22)
            UiAlign("right")
            UiText(horn.hornkey)
            UiTranslate(10, 0)
            UiFont("regular.ttf", 22)
            UiAlign("left")
            UiText("Horn")
        UiPop()
    end
    
    --speedometer above vh hp
    if (global.speedmtr_enabled) and (global.playervh ~= 0) then
        UiPush()
		    UiFont("bold.ttf", 20)
            UiTranslate(UiCenter(), UiHeight()-82)
            local maxspd = speedmtr.lim
            local mult = 3.6
            local unit = " KM/H"
            if not speedmtr.kph then
                maxspd = maxspd * 0.6214
                mult = 2.23694
                unit = " MPH"
            end
            local airspeed = getBodySpeed(GetVehicleBody(global.playervh))*mult
            local speed = airspeed/maxspd
            if speed > 1 then speed = 1 end
            UiTranslate(-100, 0)
            progressBar(200, 20, speed)
            UiTranslate(100, -12)
            if speed > 0.9 then
                if math.mod(GetTime(), 1.0) < 0.5 then
                    UiColor(1, 0, 0,  1.0)
                else
                    UiColor(1, 0, 0,  0.1)
                end
            elseif speed > 0.5 then
                UiColor(1, 0, 0)
            end
            UiAlign("center middle")
		    UiText(round(airspeed, 0)..unit)
	    UiPop()
    end

    --Sidebar for pause physics and flight
    if global.sidebar then
        local wAmnt = 1
        local ftb
        local offset = 15

        local DispList = {
            {title="Artzert's Utilities", icon="MOD/image/a-utils_smallest.png"}
        }
        

        if global.multiplayer then--TDMP comp
            ftb = {}
            local append = " (client)"
            if global.ishost then
                append = " (host)"
            end
            ftb.title = "TDMP v"..TDMP_Version()..append
            ftb.icon = "MOD/image/tdmp_smallest.png"
            table.insert(DispList, ftb)
            wAmnt = 2
            offset = 25
        end

        if global.fly_enabled then
            --flight
            ftb = {}
            ftb.title = fly.state:gsub("^%l", string.upper)
            ftb.icon = "MOD/image/flight_smallest.png"
            table.insert(DispList, ftb)
        end
        
        if global.phys_enabled then
            --pause physics
            ftb = {}
            ftb.title = "Realtime"
            ftb.icon = "MOD/image/play_smallest.png"
            if not physics.state then
                ftb.icon = "MOD/image/pausephysics_smallest.png"
                ftb.title = "Paused"
            end
            table.insert(DispList, ftb)
        end

        if global.debris_enabled then
            --debris
            ftb = {}
            ftb.title = debris.count .." Removed"
            ftb.icon = "MOD/image/debris_smallest.png"
            table.insert(DispList, ftb)
        end

        if global.tctrl_enabled then
            ftb = {}
            local cur = round(tctrl.cur, 2)
            local status = "Inactive: "
            if tctrl.active then
                status = "Active: "
            end
            ftb.title = status.." x" .. cur
            if tctrl.active then
                ftb.icon = "MOD/image/clock_smallest.png"
            else
                ftb.icon = "MOD/image/power_smallest.png"
            end
            table.insert(DispList, ftb)
        end
    
        UiPush()
            UiTranslate(15,offset)
            for i,v in pairs(DispList) do
                UiAlign("left top")
                if i>wAmnt then
                    UiColor(0,0,0,0.6)
                else
                    UiColor(1,1,1,0.6)
                end
                UiImageBox("ui/common/box-solid-6.png", 160, 26, 10, 10)
                UiPush()
                    if i>wAmnt then
                        UiColor(1,1,1,0.6)
                    else
                        UiColor(0,0,0,0.6)
                    end
                    UiAlign("left middle")
                    UiTranslate(8,13)
                    UiImage(v.icon)
                    UiFont("regular.ttf", 15)
                    UiTranslate(18,0)
                    UiText(v.title)
                UiPop()
                UiTranslate(0,31)
            end
        UiPop()
    end

    --car keys car overlay
    if global.advengine_enabled then
        if GetString("game.player.tool") == "carkeys" then
            local dir = TransformToParentVec(GetCameraTransform(), Vec(0, 0, -1))
            local hit, dist, normal, shape = QueryRaycast(GetCameraTransform().pos, dir, 100)
            if hit then
                local bd = GetShapeBody(shape)
                local vh = GetBodyVehicle(bd)
                if vh ~= 0 then
                    DrawBodyOutline(bd, 1, 1, 1, (math.sin(5*GetTime())/2+1))
                    if InputReleased("lmb") then
                        if advengine.vehicles[vh] == nil then
                            advengine.vehicles[vh] = {true, advengine.doautopark}
                        else
                            advengine.vehicles[vh][1] = not advengine.vehicles[vh][1]
                        end
                        PlaySound(advengine.beepbeep)
                        local ind = find_value(calarm.checkable, vh) 
                        if (ind ~= false) then
                            vhdata.mxhp[vh] = GetVhVox(vh)
                            calarm.state[vh] = false
                        end
                    end
                end
            end
        end
    end
end